/* very simple app to set the tv frequency */

#include <sys/time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <linux/fs.h>
#include <linux/videodev2.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

#define	MY_WIDTH	704
#define	MY_HEIGHT	480

#define max(a,b) ((a) > (b) ? (a) : (b))

#define MAXSTDS		10
#define	MAXINPS		10

typedef struct {
	Widget			w_item;
	struct v4l2_standard	std;
} StdMenuItem;

typedef struct {
	Widget			w_item;
	struct	v4l2_input	inp;
} InpMenuItem;


XtAppContext	app;
Widget		w_toplevel;
Widget		w_topbox;
Widget		w_menubox;
Widget		w_quit;
Widget		w_viewport;
Widget		w_inpbutton;
Widget		w_inpmenu;
Widget		w_stdbutton;
Widget		w_stdmenu;
Widget		w_sizebutton;
Widget		w_sizemenu;
GC		imagegc;
int		x_depth;
struct v4l2_format	fmt;

InpMenuItem	inpitem[MAXINPS];
StdMenuItem	stditem[MAXSTDS];

char		*device = "/dev/video";
int		vid;

struct { int width, height; } capturesize[] = 
{
	{160, 120},
	{192, 144},
	{320, 240},
	{384, 288},
	{640, 480},
	{704, 480}, 
	{768, 576},
};
void	inp_menu_callback(Widget w, XtPointer client, XtPointer call);
void	std_menu_callback(Widget w, XtPointer client, XtPointer call);
void	size_menu_callback(Widget w, XtPointer client, XtPointer call);
void	resize_it(Widget w, XtPointer closure, XEvent *event, Boolean *conture_dispatch);
void	quit(Widget w, XtPointer client, XtPointer call);
void	done(void);
void	stop_capturing(void);
int	start_capturing(void);
long	frame_buffer_format(Display *d, Window w);

int 
main(int argc,char **argv)
{
	int	i;
	int	err;

	atexit(done);

	if ((vid = open(device, O_RDONLY)) < 0) {
		perror(device);
		exit(1);
	}

	w_toplevel = XtAppInitialize(&app, "preview", NULL, 0,
		&argc, argv, NULL, NULL, 0);
	w_topbox = XtVaCreateManagedWidget(
		"topbox",
		boxWidgetClass, w_toplevel,
		XtNorientation, XtorientVertical,
		XtNborderWidth, 0,
		XtNhSpace, 0,
		XtNvSpace, 0,
		NULL);
	w_menubox = XtVaCreateManagedWidget(
		"menubox",
		boxWidgetClass, w_topbox,
		XtNorientation, XtorientHorizontal,
		XtNborderWidth, 0,
		XtNhSpace, 4,
		XtNvSpace, 1,
		NULL);
	w_inpbutton = XtVaCreateManagedWidget(
		"menubutton",
		menuButtonWidgetClass, w_menubox,
		XtNlabel, "Source",
		XtNmenuName, "source",
		XtNborderWidth, 0,
		NULL);
	w_inpmenu = XtVaCreatePopupShell(
		"source", /* same as XtNmenuName and static */
		simpleMenuWidgetClass, w_inpbutton,
		NULL);
	for (i = 0; ; ++i)
	{
		struct v4l2_input	einp;
		einp.index = i;
		err = ioctl(vid, VIDIOC_ENUMINPUT, &einp);
		if (err)
			break;
		inpitem[i].inp = einp;
		inpitem[i].w_item = XtVaCreateManagedWidget(
			einp.name,
			smeBSBObjectClass, w_inpmenu,
			XtNlabel, einp.name,
			NULL);
		XtAddCallback(inpitem[i].w_item,
			      XtNcallback, inp_menu_callback,
			      &inpitem[i]);
	}

	w_stdbutton = XtVaCreateManagedWidget(
		"menubutton",
		menuButtonWidgetClass, w_menubox,
		XtNlabel, "Standard",
		XtNmenuName, "standard",
		XtNborderWidth, 0,
		NULL);
	w_stdmenu = XtVaCreatePopupShell(
		"standard", /* same as XtNmenuName and static */
		simpleMenuWidgetClass, w_stdbutton,
		NULL);
	for (i = 0; ; ++i)
	{
		struct v4l2_enumstd	estd;
		estd.index = i;
		err = ioctl(vid, VIDIOC_ENUMSTD, &estd);
		if (err)
			break;
		stditem[i].std = estd.std;
		stditem[i].w_item = XtVaCreateManagedWidget(
			estd.std.name,
			smeBSBObjectClass, w_stdmenu,
			XtNlabel, estd.std.name,
			NULL);
		XtAddCallback(stditem[i].w_item,
			      XtNcallback, std_menu_callback,
			      &stditem[i]);
	}

	w_sizebutton = XtVaCreateManagedWidget(
		"menubutton",
		menuButtonWidgetClass, w_menubox,
		XtNlabel, "Size",
		XtNmenuName, "size",
		XtNborderWidth, 0,
		NULL);
	w_sizemenu = XtVaCreatePopupShell(
		"size", /* same as XtNmenuName and static */
		simpleMenuWidgetClass, w_sizebutton,
		NULL);
	for (i = 0; i < sizeof(capturesize)/sizeof(capturesize[0]); ++i)
	{
		char	sizestr[32];
		Widget	w;

		sprintf(sizestr, "%d x %d", capturesize[i].width,
			capturesize[i].height);
		w = XtVaCreateManagedWidget(
			sizestr,
			smeBSBObjectClass, w_sizemenu,
			XtNlabel, sizestr,
			NULL);
		XtAddCallback(w,
			      XtNcallback, size_menu_callback,
			      (char *)i);
	}

	w_quit = XtVaCreateManagedWidget(
		"Quit!",
		commandWidgetClass, w_menubox,
		XtNborderWidth, 0,
		NULL);
	XtAddCallback(w_quit, XtNcallback, quit, NULL);
	w_viewport = XtVaCreateManagedWidget(
		"port",
		portholeWidgetClass, w_topbox,
		XtNwidth, MY_WIDTH,
		XtNheight, MY_HEIGHT,
		NULL);
	XtRealizeWidget(w_toplevel);
	imagegc = XCreateGC(
		XtDisplay(w_viewport),
		XtWindow(w_viewport),
		0, NULL);


	XtAddEventHandler(w_toplevel,
		StructureNotifyMask | SubstructureNotifyMask, 
		False, resize_it, &w_viewport);

	for (;;) {
		XEvent event;

		err = 0;
		while (XtAppPending(app))
		{
			XtAppNextEvent(app, &event);
			switch (event.type)
			{
			default:
				break;
			}
			XtDispatchEvent(&event);
		}
	}
}


void
inp_menu_callback(Widget w, XtPointer client, XtPointer call)
{
	InpMenuItem	*smi = (InpMenuItem *)client;

	stop_capturing();

	if (ioctl(vid, VIDIOC_S_INPUT, &smi->inp.index) < 0) {
		perror("VIDIOC_S_INPUT");
	}
	start_capturing();
}

void
std_menu_callback(Widget w, XtPointer client, XtPointer call)
{
	StdMenuItem	*smi = (StdMenuItem *)client;

	stop_capturing();
	if (ioctl(vid, VIDIOC_S_STD, &smi->std) < 0) {
		perror("VIDIOC_S_STD");
	}
	start_capturing();
}

void
size_menu_callback(Widget w, XtPointer client, XtPointer call)
{
	int	i = (int)client;

	fmt.fmt.pix.width = capturesize[i].width;
	fmt.fmt.pix.height = capturesize[i].height;

	stop_capturing();

#ifdef	NEVER
	if (ioctl(vid, VIDIOC_S_FMT, &fmt) < 0) {
		perror("VIDIOC_S_FMT");
	}
#endif

	start_capturing();
}

void 
quit(Widget w, XtPointer client, XtPointer call)
{
	exit(0);
}

void
resize_it(Widget w, XtPointer closure, XEvent *event, Boolean *conture_dispatch)
{
	stop_capturing();
	start_capturing();
}

void
done(void)
{
	int	i;

	if (vid < 1)
		return;

	i = 0;
	if (ioctl(vid, VIDIOC_PREVIEW, &i) < 0) {
		perror("VIDIOC_PREVIEW");
	}

	close(vid);
}


void 
stop_capturing(void)
{
}



int
start_capturing(void)
{
	struct v4l2_window 		win;
	int	i;
	Arg	arg[10];
	Dimension	x;
	Dimension	x_view;
	Dimension	y;
	Dimension	y_view;
	Dimension	width;
	Dimension	height;

#ifdef NEVER
	fmt.fmt.pix.pixelformat = frame_buffer_format(XtDisplay(w_viewport),
		DefaultRootWindow(XtDisplay(w_viewport)));

	if (ioctl(vid, VIDIOC_S_FMT, &fmt) < 0) {
		perror("VIDIOC_S_FMT");
		return 1;
	}
#endif
x=-1;
y=-1;
width=-1;
height=-1;

	XtSetArg(arg[0], XtNx, &x);
	XtSetArg(arg[1], XtNy, &y);
	XtSetArg(arg[2], XtNwidth, &width);
	XtSetArg(arg[3], XtNheight, &height);

	XtGetValues(w_viewport, arg, 4);

	x_view = x;
	y_view = y;

	XtSetArg(arg[0], XtNx, &x);
	XtSetArg(arg[1], XtNy, &y);
	XtSetArg(arg[2], XtNwidth, &width);
	XtSetArg(arg[3], XtNheight, &height);

	XtGetValues(w_toplevel, arg, 4);

	x_view += x;
	y_view += y;

	XtSetArg(arg[0], XtNwidth, max(160, fmt.fmt.pix.width));
	XtSetArg(arg[1], XtNheight, max(120, fmt.fmt.pix.height));
	XtSetArg(arg[2], XtNbackground, 0);
	XtSetValues(w_viewport, arg, 3);
	XtRealizeWidget(w_toplevel);

	win.x = x_view;
	win.y = y_view;
	win.width = max(160,fmt.fmt.pix.width);
	win.height = max(120,fmt.fmt.pix.height);
	win.chromakey = -1;
	win.clips = NULL;
	win.clipcount = 0;

	i = 1;
	if (ioctl(vid, VIDIOC_PREVIEW, &i) < 0) {
		perror("VIDIOC_PREVIEW");
	}

	if (ioctl(vid,VIDIOC_S_WIN,&win) < 0) {
		perror("VIDIOC_S_WIN");
	}

	return 0;
}

long
frame_buffer_format(Display *d, Window w)
{
	__u32	fbf = V4L2_PIX_FMT_RGB565;

	if (x_depth == 0)
	{
#ifdef X_DEPTH
		x_depth = X_DEPTH;
#else
		XWindowAttributes wa;
		XGetWindowAttributes(d, w, &wa);
		x_depth = wa.depth;
#endif
	}
	if (x_depth == 15)
		fbf = V4L2_PIX_FMT_RGB555;
	else if (x_depth == 16)
		fbf = V4L2_PIX_FMT_RGB565;
	else if (x_depth == 24)
		fbf = V4L2_PIX_FMT_BGR24;
	else if (x_depth == 32)
		fbf = V4L2_PIX_FMT_BGR32;
	else
	{
		printf("Unrecognized display depth. You may get an "
		       "X Windows error.\n");
	}
	return fbf;
}
